from typing import Optional, Any, Union, List
import os
from dotenv import load_dotenv

import openai
from aitemplates.oai.ApiManager import SingleApiManager

from aitemplates.oai.utils.wrappers import retry_openai_api
from aitemplates.oai.types.chat import (
    ChatSequence,
    ChatConversation,
    Message,
)
from aitemplates.oai.types.functions import FunctionDef, Functions

dotenv_path = os.path.join(
    os.getcwd(), ".env"
)  # get the path to .env file in current working directory
load_dotenv(dotenv_path)  # load environment variables from the .env file

OPENAI_API_KEY = os.getenv("OPENAI_API_KEY")
model = os.getenv("model")

if OPENAI_API_KEY is None:
    raise Exception("API key not found in environment variables")

openai.api_key = OPENAI_API_KEY


@retry_openai_api()
def create_chat_completion(
    messages: Union[ChatSequence, ChatConversation, Message, List[Message], str],
    model: str = model or "gpt-3.5-turbo-0613",
    temperature: Optional[float] = 0,
    max_tokens: Optional[int] = None,
    top_p: Optional[float] = 1,
    n: Optional[int] = 1,
    stop: Optional[str] = None,
    presence_penalty: Optional[float] = 0,
    frequency_penalty: Optional[float] = 0,
    functions: Optional[Functions] | Optional[list[FunctionDef]] = None,
    function_call: Optional[object] = None,
    send_object: bool = False,
    auto_call_func: bool = False,
) -> Any:
    """Create a chat completion using the OpenAI API

    Args:
        messages (list[MessageDict]): The messages to send to the chat completion.
        model (str, optional): The model to use. Defaults to "gpt-3.5-turbo".
        temperature (float, optional): The temperature to use. Defaults to 0.
        max_tokens (int, optional): The maximum tokens to use. Defaults to None.
        top_p (float, optional): The nucleus sampling probability. Defaults to 1.
        n (int, optional): The number of messages to generate. Defaults to 1.
        stop (str, optional): The sequence at which the generation will stop. Defaults to None.
        presence_penalty (float, optional): The presence penalty to use. Defaults to 0.
        frequency_penalty (float, optional): The frequency penalty to use. Defaults to 0.
        functions (Optional[Functions], optional): The functions to use. Defaults to None.
        send_object (bool, optional): Whether to return the response object. Defaults to False.

    Returns:
        Any: The response from the chat completion.
    """
    kwarg_messages = None
    
    if isinstance(messages, str):
        messages = Message("system", messages)
    elif isinstance(messages, ChatConversation): # we set it to the last sequence which the response is None
        kwarg_messages = messages.conversation_history[-1].prompt.raw()
    elif isinstance(messages, Message):
        kwarg_messages = ChatSequence([messages]).raw()
        messages = ChatSequence([messages])
    elif isinstance(messages, List):
        kwarg_messages = ChatSequence(messages).raw()
        messages = ChatSequence(messages)
        kwarg_messages = messages.raw()
    else:
        kwarg_messages = messages.raw()

    api_manager = SingleApiManager()
    kwargs = {
        "model": model,
        "messages": kwarg_messages,
        "temperature": temperature,
        "max_tokens": max_tokens,
        "top_p": top_p,
        "n": n,
        "stop": stop,
        "presence_penalty": presence_penalty,
        "frequency_penalty": frequency_penalty,
    }

    function_pairs = None
    
    # if it's a ChatSequence being passed in
    if hasattr(messages, 'function_pairs'):
        kwargs["functions"] = messages.function_pairs.get_function_defs(dict=True)
        function_pairs = messages.function_pairs
    elif functions:
        if isinstance(functions[0], FunctionDef):
            kwargs["functions"] = [func.__dict__ for func in functions]
        else:
            function_pairs = functions
            # if you're passing in global functions from a ChatConversation sequential call
            if "functions" in kwargs:
                # add functions not in the existing functions array
                existing_function_names = {func["name"] for func in kwargs["functions"]}
                new_functions = [
                    func
                    for func in functions.get_function_defs(dict=True)
                    if func["name"] not in existing_function_names
                ]
                kwargs["functions"].extend(new_functions)
            else:
                kwargs["functions"] = functions.get_function_defs(dict=True)

    if function_call:
        kwargs["function_call"] = function_call

    response = openai.ChatCompletion.create(**kwargs)

    function_result = None

    if response.choices[0].message.get("function_call") and auto_call_func:
        function_result = Functions.execute_function_call(
            response.choices[0].message.function_call, function_pairs
        )

    api_manager.update_cost(
        response.usage.prompt_tokens, response.usage.completion_tokens, response.model
    )

    if isinstance(messages, ChatConversation):
        if function_result:
            # we want to set the content to the function_result
            messages.conversation_history[-1].update_response(response, function_result)
        else:
            messages.conversation_history[-1].update_response(response)

    if send_object:
        if function_result:
            return response, function_result
        else:
            return response
    elif n and n > 1:
        return response.choices
    elif function_result:
        return function_result
    else:
        return response.choices[0].message["content"]



import numpy as np
from typing import Any, Union
import os
from dotenv import load_dotenv

import openai

from aitemplates.oai.utils.wrappers import retry_openai_api
from aitemplates.oai.ApiManager import SingleApiManager

dotenv_path = os.path.join(os.getcwd(), '.env')  # get the path to .env file in current working directory
load_dotenv(dotenv_path)  # load environment variables from the .env file

OPENAI_API_KEY = os.getenv("OPENAI_API_KEY")

if OPENAI_API_KEY is None:
    raise Exception("API key not found in environment variables")

openai.api_key = OPENAI_API_KEY

Embedding = Union[list[np.float32], np.ndarray[Any, np.dtype[np.float32]]]
"""Embedding vector"""
TText = list[int]
"""Token array representing text"""

@retry_openai_api()
def get_embedding(
    embed: Union[str, TText, list[str], list[TText]],
    model: str = "text-embedding-ada-002",
) -> Union[Embedding, list[Embedding]]:
    """Get an embedding from the ada model.

    Args:
        embed: Input text to get embeddings for, encoded as a string or array of tokens.
            Multiple inputs may be given as a list of strings or token arrays.
        openai_api_key: OpenAI API key.
        model: The OpenAI embedding model to use. Defaults to "text-embedding-ada-002".

    Returns:
        List[float]: The embedding.
    """
    api_manager = SingleApiManager()
    multiple = isinstance(embed, list) and all(not isinstance(i, int) for i in input)

    # clean the input string
    if isinstance(embed, str):
        embed = embed.replace("\n", " ")
    elif multiple and isinstance(embed[0], str):
        embed = [text.replace("\n", " ") for text in embed]

    embeddings = openai.Embedding.create(
        input=embed,
        model=model,
    )
    
    api_manager.update_cost(
        embeddings.usage.prompt_tokens,
        0,
        embeddings.model
    )

    if not multiple:
        return embeddings.data[0]["embedding"]

    # sort the multiple return embeddings in their correct order
    embeddings = sorted(embeddings.data, key=lambda x: x["index"])
    return [d["embedding"] for d in embeddings.data]



from __future__ import annotations


from dataclasses import dataclass
from typing import List, Union, Tuple, Any, Callable, cast, Optional
from aitemplates.oai.types.base import FunctionCall

from jsonschema import validate

@dataclass
class FunctionDef:
    """Represents a "function" in OpenAI, which is mapped to a Command in Auto-GPT"""

    name: str
    description: str
    parameters: dict[str, ParameterSpec]

    @dataclass
    class ParameterSpec:
        name: str
        type: str
        description: Optional[str]
        required: bool = False

    @property
    def __dict__(self):
        """Output an OpenAI-consumable function specification"""
        return {
            "name": self.name,
            "description": self.description,
            "parameters": {
                "type": "object",
                "properties": {
                    param.name: {
                        "type": param.type,
                        "description": param.description,
                    }
                    for param in self.parameters.values()
                },
                "required": [
                    param.name for param in self.parameters.values() if param.required
                ],
            },
        }
    
class FunctionPair:
    function_pairs: Tuple[FunctionDef, Callable[..., Any]]
    
    def __init__(self, function_def: FunctionDef, function: Callable[..., Any]) -> None:
        self.function_pairs = (function_def, function)
    
    def __iter__(self):
        return iter(self.function_pairs)
    
    @property
    def definition(self) -> object:
        return self.function_pairs[0]
    
    @property
    def function(self) -> Callable[..., Any]:
        return self.function_pairs[1]
    
    def call(self, *args, **kwargs) -> Any:
        return self.function.call(*args, **kwargs)

    def update__def(self, new_def: FunctionDef) -> 'FunctionPair':
        self.function_pairs = (new_def, self.function_pairs[1])
        return self

    def update_function(self, new_function: Callable[..., Any]) -> 'FunctionPair':
        self.function_pairs = (self.function_pairs[0], new_function)
        return self
   
@dataclass
class Functions:
    function_pairs: List[FunctionPair]

    def __getitem__(self, i: int) -> FunctionPair:
        return self.function_pairs[i]

    def __iter__(self):
        return iter(self.function_pairs)

    def __len__(self) -> int:
        return len(self.function_pairs)

    def __add__(self, other: 'Functions') -> 'Functions':
        return Functions(self.function_pairs + other.function_pairs)
    
    def get_function_defs(self, dict: bool = False) -> List[object]:
        if dict:
            return [fp.definition.__dict__ for fp in self.function_pairs]
        return [fp.definition for fp in self.function_pairs]
    
    def get_functions(self) -> List[Callable[..., Any]]:
        return [cast(Callable[..., Any], function) for _, function in self.function_pairs]

    def add_function_pairs(self, functions: Union[FunctionPair, List[FunctionPair], 'Functions']):
        if isinstance(functions, Functions):
            self.function_pairs.extend(functions.function_pairs)
        elif isinstance(functions, list):
            self.function_pairs.extend(functions)
        else:
            self.function_pairs.append(functions)

    def set_function_pairs(self, functions: Union[FunctionPair, List[FunctionPair], 'Functions']):
        if isinstance(functions, Functions):
            self.function_pairs = functions.function_pairs
        elif isinstance(functions, list):
            self.function_pairs = functions
        else:
            self.function_pairs = [functions]
    
    def append(self, function_pair: FunctionPair):
        return self.function_pairs.append(function_pair)

    def extend(self, function_pairs: List[FunctionPair] | Functions):
        if isinstance(function_pairs, Functions):
            self.function_pairs.extend(function_pairs.function_pairs)
        else:
            self.function_pairs.extend(function_pairs)
    
    @staticmethod
    def execute_function_call(message: FunctionCall, functions: 'Functions') -> Any:
        function_name = message.name
        function_args = eval(message.arguments)
        
        # get the correct function
        target_function = next((fp.function for fp in functions.function_pairs if fp.definition.__dict__["name"] == function_name), None)
        
        if target_function:
            results = target_function(**function_args)
        else:
            results = f"Error: function {function_name} does not exist"
            
        return results


from __future__ import annotations

from unittest.mock import patch
import functools
import time
import os
import logging


from openai.error import (
    APIError,
    RateLimitError,
    InvalidRequestError,
    APIConnectionError,
    Timeout,
)


def retry_openai_api(
    num_retries: int = 10,
    backoff_base: float = 2.0,
    warn_user: bool = True,
):
    """Decorate a function to retry OpenAI API call when it fails.

    This function uses exponential backoff strategy for retries.

    Args:
        num_retries (int, optional): Number of retries. Defaults to 10.
        backoff_base (float, optional): Base for exponential backoff. Defaults to 2.0.
        warn_user (bool, optional): Whether to warn the user. Defaults to True.

    Returns:
        Callable: The decorated function.
    """

    def _wrapper(func):
        @functools.wraps(func)
        def _wrapped(*args, **kwargs):
            if "OPENAI_API_KEY" not in os.environ:
                raise ValueError(
                    "OPENAI_API_KEY environment variable must be set when using OpenAI API."
                )
            user_warned = not warn_user
            num_attempts = num_retries + 1  # +1 for the first attempt
            for attempt in range(1, num_attempts + 1):
                try:
                    return func(*args, **kwargs)

                except RateLimitError as e:
                    if attempt == num_attempts:
                        raise

                    logging.error("Error: Reached rate limit, passing...")
                    if not user_warned:
                        logging.warning(
                            """Please double check that you have setup a paid OpenAI API 
                              Account. You can read more here: https://docs.agpt.co/setup/#getting-an-api-key"""
                        )
                        user_warned = True
                except InvalidRequestError as e:
                    if 'not exist' in str(e):
                        logging.warning(f"Requested model does not exist. Using default model gpt-3.5-turbo-0613.")
                        kwargs['model'] = "gpt-3.5-turbo-0613"
                        return func(*args, **kwargs)
                    else:
                        logging.error(f"OpenAI API Invalid Request: {e}")
                        user_warned = True
                except APIConnectionError as e:
                    logging.error(
                        f"OpenAI API Connection Error: {e}"
                    )
                    user_warned = True
                except Timeout as e:
                    logging.error(f"OpenAI API Timeout Error: {e}")
                    user_warned = True
                except APIError as e:
                    if (e.http_status not in [502, 429]) or (attempt == num_attempts):
                        raise

                backoff = backoff_base ** (attempt + 2)
                logging.error(f"Error: API Bad gateway. Waiting {backoff} seconds...")
                time.sleep(backoff)

        return _wrapped

    return _wrapper



from __future__ import annotations

from dataclasses import dataclass
import numpy as np
from typing import Literal, TypedDict, Optional, Union, Any

# Embedding vector
Embedding = Union[list[np.float32], np.ndarray[Any, np.dtype[np.float32]]]

# Token array representing text
TText = list[int]

# Different roles the chat API takes
MessageRole = Literal["system", "user", "assistant", "function"]


class MessageDict(TypedDict):
    role: MessageRole
    content: str

@dataclass
class ResponseDict:
    index: int
    message: Message
    finish_reason: str
    model: str
    created: str
    
    def __init__(self, response: dict, function_response: Any = None):
        choice = response["choices"][0]  # Getting the first choice
        self.index = choice["index"]
        if function_response:
            self.message = Message("function", function_response)
        else:
            self.message = Message(**choice["message"])
        self.finish_reason = choice["finish_reason"]
        self.model = response["model"]
        self.created = response["created"]
        
    def __getitem__(self, i: int):
        return self.message
    
    @staticmethod
    def convert_to_response_dict(response: Union[dict, tuple]) -> Union['ResponseDict', tuple[ResponseDict, Any]]:
        if response is not None and isinstance(response, tuple):
            # if it's a tuple that means a function was called
            return (ResponseDict(response[0]), response[1])
        else: 
            return ResponseDict(response)
    
    @staticmethod
    def convert_to_raw(response_dict: ResponseDict) -> MessageDict| tuple[MessageDict, Any]:
        if response_dict is not None and isinstance(response_dict, tuple):
            # if it's a tuple that means a function was called
            return (response_dict.message.raw(), response_dict[1])
        else: 
            return response_dict.message.raw()
        
        


class UsageDict(TypedDict):
    prompt_tokens: Optional[int]
    completion_tokens: Optional[int]
    total_tokens: int


@dataclass
class FunctionCall:
    name: str
    arguments: str

@dataclass
class Message:
    """OpenAI Message object containing a role and the message content"""

    role: MessageRole
    content: str
    function_call: Optional[FunctionCall] = None

    def raw(self) -> MessageDict:
        return {"role": self.role, "content": self.content}


@dataclass
class ModelInfo:
    """Struct for model information."""

    name: str
    prompt_token_cost: float
    completion_token_cost: float
    max_tokens: int


@dataclass
class ChatModelInfo(ModelInfo):
    "Struct for chat model information."


@dataclass
class EmbeddingModelInfo(ModelInfo):
    "Struct for embedding model information."

    embedding_dimensions: int